from datetime import (
    date,
)

from django.db import (
    models,
)
from django.db.models import (
    UUIDField,
)
from lxml import (
    etree,
)

from m3_gar.importer.table import (
    BadTableError,
)
from m3_gar.importer.table.table import (
    Table,
    TableIterator,
)


class XMLIterator(TableIterator):

    def __init__(self, table, fd):
        super().__init__(table)
        self._fd = fd
        self._context = etree.iterparse(self._fd)

    def format_row(self, row):
        for key, value in row.items():
            key = key.lower()
            field = self.model._meta.get_field(key)

            if isinstance(field, UUIDField):
                value = value or None
            elif isinstance(field, models.DateField):
                value = date.fromisoformat(value) if value else None
            elif field.one_to_one or field.many_to_one:
                key = field.get_attname()
                if isinstance(field.target_field, models.IntegerField):
                    value = int(value) if value else None
                if field.null and (not value or value == '0'):
                    value = None
            elif isinstance(field, models.BooleanField):
                value = value in ['1', 'true', True]
            elif isinstance(field, models.IntegerField):
                value = int(value) if value else None
            else:
                value = value.strip().replace('\n', '').replace('\r', '')

            yield key, value

    def get_next(self):
        event, row = next(self._context)

        if row.getparent() is None:
            raise StopIteration

        item = self.process_row(row)
        row.clear()
        while row.getprevious() is not None:
            del row.getparent()[0]

        return item


class XMLTable(Table):
    iterator = XMLIterator

    def rows(self, tablelist):
        xml = self.open(tablelist=tablelist)

        try:
            return self.iterator(self, xml)
        except etree.XMLSyntaxError as e:
            raise BadTableError(
                f'Error occured during opening table `{self.name}`: {e}'
            )



import datetime
import os
import shutil
from pathlib import (
    Path,
)


class SourceWrapper:
    source = None

    def __init__(self, source, **kwargs):
        pass

    def get_date_info(self, filename):
        date_str = filename.split('_')[-2]
        return datetime.datetime.strptime(date_str, '%Y%m%d').date()

    def get_file_list(self):
        raise NotImplementedError()

    def open(self, filename):
        raise NotImplementedError()

    def __getstate__(self):
        state = self.__dict__.copy()

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)


class DirectoryWrapper(SourceWrapper):
    is_temporary = False

    def __init__(self, source, is_temporary=False, **kwargs):
        super().__init__(source=source, **kwargs)
        self.is_temporary = is_temporary
        self.source = os.path.abspath(source)

    def get_file_list(self):
        source = Path(self.source)

        return [
            str(filepath.relative_to(source))
            for filepath
            in source.glob('**/*')
            if filepath.is_file()
        ]

    def get_full_path(self, filename):
        return os.path.join(self.source, filename)

    def open(self, filename):
        return open(self.get_full_path(filename), 'rb')

    def __del__(self):
        if self.is_temporary:
            shutil.rmtree(self.source, ignore_errors=True)


class LocalArchiveWrapper(SourceWrapper):

    def __init__(self, source, **kwargs):
        super().__init__(source=source, **kwargs)
        self.source = source

    def get_file_list(self):
        return self.source.namelist()

    def open(self, filename):
        return self.source.open(filename)

    def __getstate__(self):
        state = super().__getstate__()

        state['source'] = (self.source.__class__, self.source.filename)

        return state

    def __setstate__(self, state):
        source_class, source_filename = state['source']
        state['source'] = source_class.open_file(source_filename)

        super().__setstate__(state)



from django.db import (
    models,
)


__all__ = ['Version']


class VersionManager(models.Manager):

    def nearest_by_date(self, date):
        return self.get_queryset().filter(dumpdate__lte=date).latest('dumpdate')


class Version(models.Model):

    objects = VersionManager()

    ver = models.IntegerField(primary_key=True)
    date = models.DateField(db_index=True, blank=True, null=True)
    dumpdate = models.DateField(db_index=True)

    complete_xml_url = models.CharField(max_length=255)
    delta_xml_url = models.CharField(max_length=255, blank=True, null=True)

    processed = models.BooleanField(verbose_name='Обработано', default=False)

    def __str__(self):
        return f'{self.ver} from {self.dumpdate}'



import codecs
from pathlib import (
    Path,
)

from setuptools import (
    find_packages,
)

from m3_gar.version import (
    __version__,
)


def read(fn):
    return codecs.open(Path(__file__).resolve().parent / fn).read()


setup_kwargs = dict(
    name='m3-gar',
    version=__version__,
    author='BARS Group',
    author_email='bars@bars.group',
    description='GAR Django integration for m3',
    long_description=read('README.rst'),
    license='MIT license',
    install_requires=[
        'Django>=3.2',
        'requests',
        'lxml',
        'progress',
        'gitpython',
        'twine',
        'asyncpg',
        'uvloop',
        'django-cte==1.1.5',
        'm3-gar-constants==1.0.3',
    ],
    packages=find_packages(exclude=['tests', 'test_project']),
    include_package_data=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: Russian',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



from m3_gar.config import (
    DATABASE_ALIAS,
)


class GARRouter:

    def db_for_read(self, model, **hints):
        if model._meta.app_label == 'm3_gar':
            return DATABASE_ALIAS

    def db_for_write(self, model, **hints):
        if model._meta.app_label == 'm3_gar':
            return DATABASE_ALIAS

    def allow_relation(self, obj1, obj2, **hints):
        """
        Разрешить связи из других бд к таблицам ГАР
        но запретить ссылаться из бд ГАР в другие БД
        """

        if obj1._meta.app_label == 'm3_gar' and obj2._meta.app_label == 'm3_gar':
            return True

    def allow_migrate(self, db, app_label, **hints):
        """Разрешить синхронизацию моделей в базе ГАР"""
        if app_label == 'm3_gar':
            return db == DATABASE_ALIAS
        elif db == DATABASE_ALIAS:
            return False



# Generated by Django 3.2.8 on 2021-11-05 09:37

import django.contrib.postgres.indexes
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('m3_gar', '0005_auto_20211029_1212'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='houses',
            index=django.contrib.postgres.indexes.HashIndex(fields=['objectguid'], name='m3_gar_hous_objectg_c8c379_hash'),
        ),
    ]



# This is an auto-generated Django model module.
# You'll have to do the following manually to clean this up:
#   * Rearrange models' order
#   * Make sure each model has one field with primary_key=True
#   * Make sure each ForeignKey and OneToOneField has `on_delete` set to the desired behavior
#   * Remove `managed = False` lines if you wish to allow Django to create, modify, and delete the table
# Feel free to rename the models, but don't rename db_table values or field names.
from django.db import models


class AddrObj(models.Model):
    id = models.BigIntegerField(primary_key=True, verbose_name='Уникальный идентификатор записи. Ключевое поле')
    objectid = models.BigIntegerField(verbose_name='Глобальный уникальный идентификатор адресного объекта типа INTEGER')
    objectguid = models.CharField(max_length=36, verbose_name='Глобальный уникальный идентификатор адресного объекта типа UUID')
    changeid = models.BigIntegerField(verbose_name='ID изменившей транзакции')
    name = models.CharField(max_length=250, verbose_name='Наименование')
    typename = models.CharField(max_length=50, verbose_name='Краткое наименование типа объекта')
    name_with_typename = models.CharField(
        max_length=300,
        verbose_name='Наименование c полным наименованием типа объекта',
        null=True,
    )
    level = models.CharField(max_length=10, verbose_name='Уровень адресного объекта')
    opertypeid = models.IntegerField(verbose_name='Статус действия над записью – причина появления записи')
    previd = models.BigIntegerField(verbose_name='Идентификатор записи связывания с предыдущей исторической записью', blank=True, null=True)
    nextid = models.BigIntegerField(verbose_name='Идентификатор записи связывания с последующей исторической записью', blank=True, null=True)
    updatedate = models.DateField(verbose_name='Дата внесения (обновления) записи')
    startdate = models.DateField(verbose_name='Начало действия записи')
    enddate = models.DateField(verbose_name='Окончание действия записи')
    isactual = models.BooleanField(verbose_name='Статус актуальности адресного объекта ФИАС')
    isactive = models.BooleanField(verbose_name='Признак действующего адресного объекта')

    class Meta:
        abstract = True


class AddrObjDivision(models.Model):
    id = models.BigIntegerField(primary_key=True, verbose_name='Уникальный идентификатор записи. Ключевое поле')
    parentid = models.BigIntegerField(verbose_name='Родительский ID')
    childid = models.BigIntegerField(verbose_name='Дочерний ID')
    changeid = models.BigIntegerField(verbose_name='ID изменившей транзакции')

    class Meta:
        abstract = True


class AddrObjTypes(models.Model):
    id = models.IntegerField(primary_key=True, verbose_name='Идентификатор записи')
    level = models.IntegerField(verbose_name='Уровень адресного объекта')
    shortname = models.CharField(max_length=50, verbose_name='Краткое наименование типа объекта')
    name = models.CharField(max_length=250, verbose_name='Полное наименование типа объекта')
    desc = models.CharField(max_length=250, verbose_name='Описание', blank=True, null=True)
    updatedate = models.DateField(verbose_name='Дата внесения (обновления) записи')
    startdate = models.DateField(verbose_name='Начало действия записи')
    enddate = models.DateField(verbose_name='Окончание действия записи')
    isactive = models.BooleanField(verbose_name='Статус активности')

    class Meta:
        abstract = True


class AdmHierarchy(models.Model):
    id = models.BigIntegerField(primary_key=True, verbose_name='Уникальный идентификатор записи. Ключевое поле')
    objectid = models.BigIntegerField(verbose_name='Глобальный уникальный идентификатор объекта')
    parentobjid = models.BigIntegerField(verbose_name='Идентификатор родительского объекта', blank=True, null=True)
    changeid = models.BigIntegerField(verbose_name='ID изменившей транзакции')
    regioncode = models.CharField(max_length=4, verbose_name='Код региона', blank=True, null=True)
    areacode = models.CharField(max_length=4, verbose_name='Код района', blank=True, null=True)
    citycode = models.CharField(max_length=4, verbose_name='Код города', blank=True, null=True)
    placecode = models.CharField(max_length=4, verbose_name='Код населенного пункта', blank=True, null=True)
    plancode = models.CharField(max_length=4, verbose_name='Код ЭПС', blank=True, null=True)
    streetcode = models.CharField(max_length=4, verbose_name='Код улицы', blank=True, null=True)
    previd = models.BigIntegerField(verbose_name='Идентификатор записи связывания с предыдущей исторической записью', blank=True, null=True)
    nextid = models.BigIntegerField(verbose_name='Идентификатор записи связывания с последующей исторической записью', blank=True, null=True)
    updatedate = models.DateField(verbose_name='Дата внесения (обновления) записи')
    startdate = models.DateField(verbose_name='Начало действия записи')
    enddate = models.DateField(verbose_name='Окончание действия записи')
    isactive = models.BooleanField(verbose_name='Признак действующего адресного объекта')
    path = models.TextField(verbose_name='Материализованный путь к объекту (полная иерархия)', blank=True, null=True)

    class Meta:
        abstract = True


class Apartments(models.Model):
    id = models.BigIntegerField(primary_key=True, verbose_name='Уникальный идентификатор записи. Ключевое поле')
    objectid = models.BigIntegerField(verbose_name='Глобальный уникальный идентификатор объекта типа INTEGER')
    objectguid = models.CharField(max_length=36, verbose_name='Глобальный уникальный идентификатор адресного объекта типа UUID')
    changeid = models.BigIntegerField(verbose_name='ID изменившей транзакции')
    number = models.CharField(max_length=50, verbose_name='Номер комнаты')
    aparttype = models.IntegerField(verbose_name='Тип комнаты')
    opertypeid = models.BigIntegerField(verbose_name='Статус действия над записью – причина появления записи')
    previd = models.BigIntegerField(verbose_name='Идентификатор записи связывания с предыдущей исторической записью', blank=True, null=True)
    nextid = models.BigIntegerField(verbose_name='Идентификатор записи связывания с последующей исторической записью', blank=True, null=True)
    updatedate = models.DateField(verbose_name='Дата внесения (обновления) записи')
    startdate = models.DateField(verbose_name='Начало действия записи')
    enddate = models.DateField(verbose_name='Окончание действия записи')
    isactual = models.BooleanField(verbose_name='Статус актуальности адресного объекта ФИАС')
    isactive = models.BooleanField(verbose_name='Признак действующего адресного объекта')

    class Meta:
        abstract = True


class ApartmentTypes(models.Model):
    id = models.IntegerField(primary_key=True, verbose_name='Идентификатор типа (ключ)')
    name = models.CharField(max_length=50, verbose_name='Наименование')
    shortname = models.CharField(max_length=50, verbose_name='Краткое наименование', blank=True, null=True)
    desc = models.CharField(max_length=250, verbose_name='Описание', blank=True, null=True)
    updatedate = models.DateField(verbose_name='Дата внесения (обновления) записи')
    startdate = models.DateField(verbose_name='Начало действия записи')
    enddate = models.DateField(verbose_name='Окончание действия записи')
    isactive = models.BooleanField(verbose_name='Статус активности')

    class Meta:
        abstract = True


class Carplaces(models.Model):
    id = models.BigIntegerField(primary_key=True, verbose_name='Уникальный идентификатор записи. Ключевое поле')
    objectid = models.BigIntegerField(verbose_name='Глобальный уникальный идентификатор объекта типа INTEGER')
    objectguid = models.CharField(max_length=36, verbose_name='Глобальный уникальный идентификатор адресного объекта типа UUID')
    changeid = models.BigIntegerField(verbose_name='ID изменившей транзакции')
    number = models.CharField(max_length=50, verbose_name='Номер машиноместа')
    opertypeid = models.IntegerField(verbose_name='Статус действия над записью – причина появления записи')
    previd = models.BigIntegerField(verbose_name='Идентификатор записи связывания с предыдущей исторической записью', blank=True, null=True)
    nextid = models.BigIntegerField(verbose_name='Идентификатор записи связывания с последующей исторической записью', blank=True, null=True)
    updatedate = models.DateField(verbose_name='Дата внесения (обновления) записи')
    startdate = models.DateField(verbose_name='Начало действия записи')
    enddate = models.DateField(verbose_name='Окончание действия записи')
    isactual = models.BooleanField(verbose_name='Статус актуальности адресного объекта ФИАС')
    isactive = models.BooleanField(verbose_name='Признак действующего адресного объекта')

    class Meta:
        abstract = True


class ChangeHistory(models.Model):
    changeid = models.BigIntegerField(primary_key=True, verbose_name='ID изменившей транзакции')
    objectid = models.BigIntegerField(verbose_name='Уникальный ID объекта')
    adrobjectid = models.CharField(max_length=36, verbose_name='Уникальный ID изменившей транзакции (GUID)')
    opertypeid = models.IntegerField(verbose_name='Тип операции')
    ndocid = models.BigIntegerField(verbose_name='ID документа', blank=True, null=True)
    changedate = models.DateField(verbose_name='Дата изменения')

    class Meta:
        abstract = True


class Houses(models.Model):
    id = models.BigIntegerField(primary_key=True, verbose_name='Уникальный идентификатор записи. Ключевое поле')
    objectid = models.BigIntegerField(verbose_name='Глобальный уникальный идентификатор объекта типа INTEGER')
    objectguid = models.CharField(max_length=36, verbose_name='Глобальный уникальный идентификатор адресного объекта типа UUID')
    changeid = models.BigIntegerField(verbose_name='ID изменившей транзакции')
    housenum = models.CharField(max_length=50, verbose_name='Основной номер дома', blank=True, null=True)
    addnum1 = models.CharField(max_length=50, verbose_name='Дополнительный номер дома 1', blank=True, null=True)
    addnum2 = models.CharField(max_length=50, verbose_name='Дополнительный номер дома 1', blank=True, null=True)
    housetype = models.IntegerField(verbose_name='Основной тип дома', blank=True, null=True)
    addtype1 = models.IntegerField(verbose_name='Дополнительный тип дома 1', blank=True, null=True)
    addtype2 = models.IntegerField(verbose_name='Дополнительный тип дома 2', blank=True, null=True)
    opertypeid = models.IntegerField(verbose_name='Статус действия над записью – причина появления записи')
    previd = models.BigIntegerField(verbose_name='Идентификатор записи связывания с предыдущей исторической записью', blank=True, null=True)
    nextid = models.BigIntegerField(verbose_name='Идентификатор записи связывания с последующей исторической записью', blank=True, null=True)
    updatedate = models.DateField(verbose_name='Дата внесения (обновления) записи')
    startdate = models.DateField(verbose_name='Начало действия записи')
    enddate = models.DateField(verbose_name='Окончание действия записи')
    isactual = models.BooleanField(verbose_name='Статус актуальности адресного объекта ФИАС')
    isactive = models.BooleanField(verbose_name='Признак действующего адресного объекта')

    class Meta:
        abstract = True


class HouseTypes(models.Model):
    id = models.IntegerField(primary_key=True, verbose_name='Идентификатор')
    name = models.CharField(max_length=50, verbose_name='Наименование')
    shortname = models.CharField(max_length=50, verbose_name='Краткое наименование', blank=True, null=True)
    desc = models.CharField(max_length=250, verbose_name='Описание', blank=True, null=True)
    updatedate = models.DateField(verbose_name='Дата внесения (обновления) записи')
    startdate = models.DateField(verbose_name='Начало действия записи')
    enddate = models.DateField(verbose_name='Окончание действия записи')
    isactive = models.BooleanField(verbose_name='Статус активности')

    class Meta:
        abstract = True


class MunHierarchy(models.Model):
    id = models.BigIntegerField(primary_key=True, verbose_name='Уникальный идентификатор записи. Ключевое поле')
    objectid = models.BigIntegerField(verbose_name='Глобальный уникальный идентификатор адресного объекта ')
    parentobjid = models.BigIntegerField(verbose_name='Идентификатор родительского объекта', blank=True, null=True)
    changeid = models.BigIntegerField(verbose_name='ID изменившей транзакции')
    oktmo = models.CharField(max_length=11, verbose_name='Код ОКТМО', blank=True, null=True)
    previd = models.BigIntegerField(verbose_name='Идентификатор записи связывания с предыдущей исторической записью', blank=True, null=True)
    nextid = models.BigIntegerField(verbose_name='Идентификатор записи связывания с последующей исторической записью', blank=True, null=True)
    updatedate = models.DateField(verbose_name='Дата внесения (обновления) записи')
    startdate = models.DateField(verbose_name='Начало действия записи')
    enddate = models.DateField(verbose_name='Окончание действия записи')
    isactive = models.BooleanField(verbose_name='Признак действующего адресного объекта')
    path = models.TextField(verbose_name='Материализованный путь к объекту (полная иерархия)', blank=True, null=True)

    class Meta:
        abstract = True


class NormativeDocs(models.Model):
    id = models.BigIntegerField(primary_key=True, verbose_name='Уникальный идентификатор документа')
    name = models.CharField(max_length=8000, verbose_name='Наименование документа')
    date = models.DateField(verbose_name='Дата документа')
    number = models.CharField(max_length=150, verbose_name='Номер документа')
    type = models.IntegerField(verbose_name='Тип документа')
    kind = models.IntegerField(verbose_name='Вид документа')
    updatedate = models.DateField(verbose_name='Дата обновления')
    orgname = models.CharField(max_length=255, verbose_name='Наименование органа создвшего нормативный документ', blank=True, null=True)
    regnum = models.CharField(max_length=100, verbose_name='Номер государственной регистрации', blank=True, null=True)
    regdate = models.DateField(verbose_name='Дата государственной регистрации', blank=True, null=True)
    accdate = models.DateField(verbose_name='Дата вступления в силу нормативного документа', blank=True, null=True)
    comment = models.CharField(max_length=8000, verbose_name='Комментарий', blank=True, null=True)

    class Meta:
        abstract = True


class NormativeDocsKinds(models.Model):
    id = models.IntegerField(primary_key=True, verbose_name='Идентификатор записи')
    name = models.CharField(max_length=500, verbose_name='Наименование')

    class Meta:
        abstract = True


class NormativeDocsTypes(models.Model):
    id = models.IntegerField(primary_key=True, verbose_name='Идентификатор записи')
    name = models.CharField(max_length=500, verbose_name='Наименование')
    startdate = models.DateField(verbose_name='Дата начала действия записи')
    enddate = models.DateField(verbose_name='Дата окончания действия записи')

    class Meta:
        abstract = True


class ObjectLevels(models.Model):
    level = models.IntegerField(primary_key=True, verbose_name='Уникальный идентификатор записи. Ключевое поле. Номер уровня объекта')
    name = models.CharField(max_length=250, verbose_name='Наименование')
    shortname = models.CharField(max_length=50, verbose_name='Краткое наименование', blank=True, null=True)
    updatedate = models.DateField(verbose_name='Дата внесения (обновления) записи')
    startdate = models.DateField(verbose_name='Начало действия записи')
    enddate = models.DateField(verbose_name='Окончание действия записи')
    isactive = models.BooleanField(verbose_name='Признак действующего адресного объекта')

    class Meta:
        abstract = True


class OperationTypes(models.Model):
    id = models.IntegerField(primary_key=True, verbose_name='Идентификатор статуса (ключ)')
    name = models.CharField(max_length=100, verbose_name='Наименование')
    shortname = models.CharField(max_length=100, verbose_name='Краткое наименование', blank=True, null=True)
    desc = models.CharField(max_length=250, verbose_name='Описание', blank=True, null=True)
    updatedate = models.DateField(verbose_name='Дата внесения (обновления) записи')
    startdate = models.DateField(verbose_name='Начало действия записи')
    enddate = models.DateField(verbose_name='Окончание действия записи')
    isactive = models.BooleanField(verbose_name='Статус активности')

    class Meta:
        abstract = True


class Param(models.Model):
    id = models.BigIntegerField(primary_key=True, verbose_name='Идентификатор записи')
    objectid = models.BigIntegerField(verbose_name='Глобальный уникальный идентификатор адресного объекта ')
    changeid = models.BigIntegerField(verbose_name='ID изменившей транзакции', blank=True, null=True)
    changeidend = models.BigIntegerField(verbose_name='ID завершившей транзакции')
    typeid = models.IntegerField(verbose_name='Тип параметра')
    value = models.CharField(max_length=8000, verbose_name='Значение параметра')
    updatedate = models.DateField(verbose_name='Дата внесения (обновления) записи')
    startdate = models.DateField(verbose_name='Дата начала действия записи')
    enddate = models.DateField(verbose_name='Дата окончания действия записи')

    class Meta:
        abstract = True


class ParamTypes(models.Model):
    id = models.IntegerField(primary_key=True, verbose_name='Идентификатор типа параметра (ключ)')
    name = models.CharField(max_length=50, verbose_name='Наименование')
    code = models.CharField(max_length=50, verbose_name='Краткое наименование')
    desc = models.CharField(max_length=120, verbose_name='Описание', blank=True, null=True)
    updatedate = models.DateField(verbose_name='Дата внесения (обновления) записи')
    startdate = models.DateField(verbose_name='Начало действия записи')
    enddate = models.DateField(verbose_name='Окончание действия записи')
    isactive = models.BooleanField(verbose_name='Статус активности')

    class Meta:
        abstract = True


class ReestrObjects(models.Model):
    objectid = models.BigIntegerField(primary_key=True, verbose_name='Уникальный идентификатор объекта')
    createdate = models.DateField(verbose_name='Дата создания')
    changeid = models.BigIntegerField(verbose_name='ID изменившей транзакции')
    levelid = models.IntegerField(verbose_name='Уровень объекта')
    updatedate = models.DateField(verbose_name='Дата обновления')
    objectguid = models.CharField(max_length=36, verbose_name='GUID объекта')
    isactive = models.BooleanField(verbose_name='Признак действующего объекта (1 - действующий, 0 - не действующий)')

    class Meta:
        abstract = True


class Rooms(models.Model):
    id = models.BigIntegerField(primary_key=True, verbose_name='Уникальный идентификатор записи. Ключевое поле')
    objectid = models.BigIntegerField(verbose_name='Глобальный уникальный идентификатор объекта типа INTEGER')
    objectguid = models.CharField(max_length=36, verbose_name='Глобальный уникальный идентификатор адресного объекта типа UUID')
    changeid = models.BigIntegerField(verbose_name='ID изменившей транзакции')
    number = models.CharField(max_length=50, verbose_name='Номер комнаты или офиса')
    roomtype = models.IntegerField(verbose_name='Тип комнаты или офиса')
    opertypeid = models.IntegerField(verbose_name='Статус действия над записью – причина появления записи')
    previd = models.BigIntegerField(verbose_name='Идентификатор записи связывания с предыдущей исторической записью', blank=True, null=True)
    nextid = models.BigIntegerField(verbose_name='Идентификатор записи связывания с последующей исторической записью', blank=True, null=True)
    updatedate = models.DateField(verbose_name='Дата внесения (обновления) записи')
    startdate = models.DateField(verbose_name='Начало действия записи')
    enddate = models.DateField(verbose_name='Окончание действия записи')
    isactual = models.BooleanField(verbose_name='Статус актуальности адресного объекта ФИАС')
    isactive = models.BooleanField(verbose_name='Признак действующего адресного объекта')

    class Meta:
        abstract = True


class RoomTypes(models.Model):
    id = models.IntegerField(primary_key=True, verbose_name='Идентификатор типа (ключ)')
    name = models.CharField(max_length=100, verbose_name='Наименование')
    shortname = models.CharField(max_length=50, verbose_name='Краткое наименование', blank=True, null=True)
    desc = models.CharField(max_length=250, verbose_name='Описание', blank=True, null=True)
    updatedate = models.DateField(verbose_name='Дата внесения (обновления) записи')
    startdate = models.DateField(verbose_name='Начало действия записи')
    enddate = models.DateField(verbose_name='Окончание действия записи')
    isactive = models.BooleanField(verbose_name='Статус активности')

    class Meta:
        abstract = True


class Steads(models.Model):
    id = models.IntegerField(primary_key=True, verbose_name='Уникальный идентификатор записи. Ключевое поле')
    objectid = models.IntegerField(verbose_name='Глобальный уникальный идентификатор объекта типа INTEGER')
    objectguid = models.CharField(max_length=36, verbose_name='Глобальный уникальный идентификатор адресного объекта типа UUID')
    changeid = models.IntegerField(verbose_name='ID изменившей транзакции')
    number = models.CharField(max_length=250, verbose_name='Номер земельного участка')
    opertypeid = models.CharField(max_length=2, verbose_name='Статус действия над записью – причина появления записи')
    previd = models.IntegerField(verbose_name='Идентификатор записи связывания с предыдущей исторической записью', blank=True, null=True)
    nextid = models.IntegerField(verbose_name='Идентификатор записи связывания с последующей исторической записью', blank=True, null=True)
    updatedate = models.DateField(verbose_name='Дата внесения (обновления) записи')
    startdate = models.DateField(verbose_name='Начало действия записи')
    enddate = models.DateField(verbose_name='Окончание действия записи')
    isactual = models.BooleanField(verbose_name='Статус актуальности адресного объекта ФИАС')
    isactive = models.BooleanField(verbose_name='Признак действующего адресного объекта')

    class Meta:
        abstract = True



from django.dispatch import (
    receiver,
)
from django.utils import (
    timezone,
)

from m3_gar.importer.signals import (
    post_download,
    post_fetch_version,
    post_import,
    post_import_table,
    post_load,
    post_unpack,
    post_update,
    pre_download,
    pre_fetch_version,
    pre_import,
    pre_import_table,
    pre_load,
    pre_unpack,
    pre_update,
)


class Timer:
    start = None
    end = None

    fetch_versions = None
    download = None
    unpack = None
    load = None

    @classmethod
    def full_reset(cls):
        cls.start = None
        cls.end = None
        cls.reset_counters()

    @classmethod
    def reset_counters(cls):
        cls.download = None
        cls.unpack = None
        cls.load = None

    @classmethod
    def init(cls):
        cls.full_reset()
        cls.start = timezone.now()


@receiver(pre_fetch_version)
def pre_fetch_version_callback(sender, **kwargs):
    pass


@receiver(post_fetch_version)
def post_fetch_version_callback(sender, **kwargs):
    time = timezone.now()
    print(f'Version info updated at {time}. Estimated time: {time - Timer.start}')


@receiver(pre_load)
def pre_load_callback(sender, src, **kwargs):
    pass


@receiver(post_load)
def post_load_callback(sender, wrapper, **kwargs):
    Timer.load = timezone.now()


@receiver(pre_download)
def pre_download_callback(sender, url, **kwargs):
    Timer.download = timezone.now()


@receiver(post_download)
def post_download_callback(sender, url, path, **kwargs):
    Timer.download = timezone.now() - Timer.download


@receiver(pre_unpack)
def pre_unpack_callback(sender, archive, **kwargs):
    Timer.unpack = timezone.now()


@receiver(post_unpack)
def post_unpack_callback(sender, archive, dst, **kwargs):
    Timer.unpack = timezone.now() - Timer.unpack


@receiver(pre_import_table)
def pre_import_table_callback(sender, table, **kwargs):
    pass


@receiver(post_import_table)
def post_import_table_callback(sender, table, **kwargs):
    pass


@receiver(pre_import)
def pre_import_callback(sender, version, **kwargs):
    print(f'Loading data v.{version.ver} started at {timezone.now()}')


@receiver(post_import)
def post_import_callback(sender, version, **kwargs):
    time = timezone.now()
    print(f'Data v.{version} loaded at {time}')
    print(
        f'Estimated time: {time - Timer.start}. '
        f'Download: {Timer.download or 0}. '
        f'Unpack: {Timer.unpack or 0}. '
        f'Import: {time - Timer.load}'
    )
    Timer.reset_counters()


@receiver(pre_update)
def pre_update_callback(sender, before, after, **kwargs):
    print(f'Updating from v.{before.ver} to v.{after.ver} started at {timezone.now()}')


@receiver(post_update)
def post_update_callback(sender, before, after, **kwargs):
    time = timezone.now()
    print(f'Data v.{before.ver} is updated to v.{after.ver} at {time}')
    print(
        f'Download: {Timer.unpack or 0}. '
        f'Unpack: {time - Timer.load}. '
        f'Import: {time - Timer.start}. '
        f'Total time: {Timer.download or 0}.'
    )
    Timer.reset_counters()



import inspect

from django.apps import (
    apps,
)
from django.core.management import (
    BaseCommand,
    CommandError,
)
from django.db import (
    models,
)

import m3_gar.base_models


class Command(BaseCommand):
    help = 'Утилита проверяет, что по всем базовым моделям созданы дочерние'

    def handle(self, **options):
        child_models = apps.get_models()

        success = True

        for name, obj in inspect.getmembers(m3_gar.base_models):
            if inspect.isclass(obj) and issubclass(obj, models.Model):
                for model in child_models:
                    if issubclass(model, obj):
                        break
                else:
                    self.stdout.write(f'Не найдена модель для {name}')
                    success = False

        if not success:
            self.stdout.write('Найдены не унаследованные базовые модели')

            raise CommandError



from m3_gar.importer.exceptions import (
    ImporterError,
)


class TableListLoadingError(ImporterError):
    pass


class NoNewVersionError(ImporterError):
    pass


class BadArchiveError(TableListLoadingError):

    def __init__(self, source, *args):
        super().__init__(
            f'Archive: `{source}` corrupted or is not archive',
            *args,
        )


class EmptyArchiveError(TableListLoadingError):
    def __init__(self, source, *args):
        super().__init__(
            f'Archive: `{source}` is empty',
            *args
        )


class RetrieveError(TableListLoadingError):
    pass



from django.conf import (
    settings,
)
from django.core.management import (
    BaseCommand,
    call_command,
)

from m3_gar.exceptions import (
    PublisherError,
)
from m3_gar.publish_tools import (
    ArchiveGARSchemaManager,
    DownloadGARSchemaManager,
    ReleasePreparer,
    Releaser,
    RepoManager,
)


class Command(BaseCommand):
    help = (
        'Основная команда для скачивания новых схем по указанному url, '
        'либо разархивирования архива БД ГАР, содержащего в себе схемы, '
        'генерации моделей по схемам, создания миграций, и выпуска пакета '
        'в случае наличия изменений'
    )
    usage_str = (
        'Usage: python manage.py gar_update_schema --url=<gar_xsd_schemas_url> '
        'OR python manage.py gar_update_schema --path=<path_to_archive_with_schemas>'
    )

    def add_arguments(self, parser):
        super().add_arguments(parser)

        parser.add_argument(
            '--url',
            action='store',
            dest='url',
            help='URL до zip-файла с xsd-схемами для построения БД',
        )
        parser.add_argument(
            '--path',
            action='store',
            dest='path',
            help='Путь до zip-файла с xsd-схемами для построения БД',
        )
        parser.add_argument(
            '--testmode',
            action='store_true',
            default=False,
            help=(
                'Тестовый режим - пуш будет сделан в отдельную ветку, '
                'пакет релизнется на тестовом PyPi. '
                'Для включения нужно добавить флаг --testmode'
            ),
        )

    def handle(self, *args, url, path, testmode):
        try:
            if url and path or not (url or path):
                raise PublisherError('Необходимо указать или url, или path!')

            self.stdout.write('Запуск обновления пакета по схемам ГАР')

            repo_manager = RepoManager(testmode)
            repo = repo_manager.repo

            self.stdout.write('1/8 Получение схем...')
            if url:
                gar_schema_manager = DownloadGARSchemaManager(url, repo)

            elif path:
                gar_schema_manager = ArchiveGARSchemaManager(path, repo)

            gar_schema_manager.process_schemas()
            if not gar_schema_manager.has_schema_changes():
                # по схемам не было изменений, прекращаем работу
                raise PublisherError('Изменения в формате ГАР не найдены')

            self.stdout.write('2/8 Генерация моделей...')
            call_command(
                'gar_generate_base_models',
                dst=settings.GAR_BASE_DIR / 'base_models.py',
            )

            self.stdout.write('3/8 Проверка моделей...')
            call_command('gar_check_models')

            self.stdout.write('4/8 Создание и применение миграций...')
            call_command('makemigrations', 'm3_gar')
            call_command('migrate', 'm3_gar')

            self.stdout.write('5/8 Добавление новых файлов в репозиторий...')
            repo_manager.index_add()

            self.stdout.write('6/8 Обновление версии и лога изменений...')
            preparer = ReleasePreparer()
            preparer.prepare()
            new_version = preparer.new_version_str

            self.stdout.write('7/8 Обновление репозитория...')
            repo_manager.save_changes(new_version)

            self.stdout.write('8/8 Релиз нового пакета...')
            Releaser(new_version, testmode).release()

            self.stdout.write('Обновление пакета по новым схемам ГАР выполнено успешно')

        except PublisherError as e:
            self.stderr.write(str(e))



from m3_gar.base_models import (
    Rooms as BaseRooms,
    RoomTypes as BaseRoomTypes,
)
from m3_gar.models.op_type import (
    OperationTypes,
)
from m3_gar.models.reestr import (
    ReestrObjects,
)
from m3_gar.models.util import (
    RegionCodeModelMixin,
    add_params,
    make_fk,
)


__all__ = ['Rooms', 'RoomTypes']


class RoomTypes(BaseRoomTypes):
    """
    Сведения по типам комнат
    """
    class Meta:
        verbose_name = 'Тип комнаты'
        verbose_name_plural = 'Типы комнат'


class Rooms(BaseRooms, RegionCodeModelMixin):
    """
    Сведения по комнатам
    """

    level = 12

    class Meta:
        verbose_name = 'Комната'
        verbose_name_plural = 'Комнаты'


make_fk(Rooms, 'roomtype', to=RoomTypes, null=True, blank=True)
make_fk(Rooms, 'opertypeid', to=OperationTypes)
make_fk(Rooms, 'objectid', to=ReestrObjects)

add_params(Rooms, 'm3_gar.RoomsParams')



# Generated by Django 3.2.6 on 2022-02-17 05:56

from django.db import migrations
from django.db.models import Max

from m3_gar.importer.commands import get_table_names


def mark_version_as_processed_by_status_table(apps, schema_editor):
    Status = apps.get_model('m3_gar', 'Status')
    Version = apps.get_model('m3_gar', 'Version')

    max_version = Status.objects.aggregate(Max('ver'))['ver__max']
    Version.objects.filter(ver__lte=max_version).update(processed=True)


def set_status_info_by_processed_version(apps, schema_editor):
    Status = apps.get_model('m3_gar', 'Status')
    Version = apps.get_model('m3_gar', 'Version')
    max_version = Version.objects.filter(processed=True).order_by('ver').last()

    if max_version:
        for table in get_table_names(None):
            status = Status(table=table, ver=max_version)
            status.save()


class Migration(migrations.Migration):

    dependencies = [
        ('m3_gar', '0007_version_processed'),
    ]

    operations = [
        migrations.RunPython(mark_version_as_processed_by_status_table, set_status_info_by_processed_version),
    ]



import re
from importlib import (
    import_module,
)
from os import (
    listdir,
)

from django.apps import (
    apps,
)
from django.conf import (
    settings,
)
from django.core.exceptions import (
    ImproperlyConfigured,
)
from django.db.backends.base.base import (
    BaseDatabaseWrapper,
)
from django.db.backends.base.introspection import (
    BaseDatabaseIntrospection,
)
from lxml import (
    etree,
)

from m3_gar.exceptions import (
    PublisherError,
)


def get_by_xpath(element, xpath):
    """
    Возвращает все объекты по xpath
    """
    return element.xpath(
        xpath,
        namespaces={'xs': "http://www.w3.org/2001/XMLSchema"}
    )


def get_first_by_xpath(element, xpath):
    """
    Возвращает первый объект по xpath
    """
    return get_by_xpath(element, xpath)[0]


class XSDField:
    """
    Поле в стуктуре XSD
    """

    # соотвествие атрибутов XSD и атрибутов объекта с конверторами
    attr_map = {
        'maxLength': ('internal_size', int),
        'length': ('internal_size', int),
    }

    # индексы полей при обращении через порядковый номер
    indexes = {
        0: 'name',
        1: 'type_code',
        2: 'display_size',
        3: 'internal_size',
        4: 'precision',
        5: 'scale',
        6: 'null_ok',
        7: 'default',
        8: 'collation',
        9: 'description',
        10: 'path',
    }

    def __init__(self, element):
        super().__init__()

        self.name = element.get('name', '').lower()
        self.type_code = None
        self.display_size = None
        self.internal_size = None
        self.precision = None
        self.scale = None
        self.null_ok = element.get('use') != 'required'
        self.default = None
        self.collation = None

        self.description = element[0][0].text.replace('\n', ' ')

        try:
            attrs = get_first_by_xpath(
                element,
                'xs:simpleType/xs:restriction[1]'
            )
        except IndexError:
            self.type_code = element.get('type')
        else:
            self.type_code = attrs.get('base')
            for attr_element in attrs:
                local_name = etree.QName(attr_element).localname
                if local_name in self.attr_map:
                    attr_name, convertor = self.attr_map[local_name]
                    setattr(
                        self,
                        attr_name,
                        convertor(
                            attr_element.get('value')
                        ),
                    )

        if self.type_code:
            self.type_code = self.type_code.replace(f'{element.prefix}:', '')

            if self.name in ['isactive', 'isactual'] and self.type_code == 'integer':
                self.type_code = 'boolean'

        else:
            self.type_code = 'string'
            self.null_ok = True
            self.default = ''

    def __getitem__(self, i):
        return getattr(self, self.indexes[i])


class XSDObject:
    """
    XSD объект
    """

    def __init__(self, folder_path, file_name):
        super().__init__()

        self.name = self.get_name(file_name)
        self.file_name = file_name
        # для совместимости с типом "Таблица" с точки зрения DataBaseEngine
        self.type = 't'

        root = etree.parse(str(folder_path / file_name)).getroot()

        elements = get_by_xpath(root, 'xs:element')
        elements_len = len(elements)

        if elements_len == 1:
            element = get_first_by_xpath(
                elements[0],
                'xs:complexType/xs:sequence/xs:element[1]',
            )

            self.description = get_first_by_xpath(
                element,
                'xs:annotation/xs:documentation[1]/text()',
            )
            self.fields = [
                XSDField(el) for el in get_first_by_xpath(element, 'xs:complexType')
            ]

        elif elements_len == 2:
            self.description = get_first_by_xpath(
                elements[0],
                'xs:complexType/xs:sequence/xs:element/xs:annotation/xs:documentation[1]/text()',
            )

            self.fields = [
                XSDField(el) for el in get_first_by_xpath(elements[1], 'xs:complexType')
            ]

    def get_name(self, file_name):
        """
        Возвращает имя по имени файла
        """
        p = re.compile("_[a-zA-Z_]+")
        name = p.search(file_name).group(0)[1:-1]

        return name


class XSDFolder:
    """
    Директория с XSD
    """

    def __init__(self, folder_path):
        super().__init__()

        self.files = {}

        files = [
            f for f in listdir(folder_path) if (folder_path / f).is_file()
        ]

        for file_ in files:
            xsd_obj = XSDObject(folder_path, file_)
            self.files[xsd_obj.name] = xsd_obj

    def get_files_list(self):
        return self.files.values()

    def get_file(self, name):
        return self.files[name]


class XSDFolderCursor:
    """
    "Указатель" для взаимодействия
    """

    def __init__(self, xsd_path):
        super().__init__()

        self.xsd_path = xsd_path
        self.folder = None

    def __enter__(self):
        self.folder = XSDFolder(self.xsd_path)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def get_files_list(self):
        """
        Возвращает перечень файлов
        """
        return self.folder.get_files_list()

    def get_file(self, name):
        """
        Возвращает файл по имени
        """
        return self.folder.get_file(name)


class XSDFolderIntrospection(BaseDatabaseIntrospection):
    """
    Анализатор директории с xsd
    """

    data_types_reverse = {
        'boolean': 'BooleanField',
        'long': 'BigIntegerField',
        'integer': 'IntegerField',
        'string': 'CharField',
        'date': 'DateField',
    }

    def get_table_list(self, cursor):
        """
        Возвращает перечень файлов
        """
        return cursor.get_files_list()

    def get_relations(self, cursor, table_name):
        """
        Возвращает перечнь связей между файлами (пока считаем, что их нет)
        """
        return {}

    def get_primary_key_column(self, cursor, table_name):
        return self.get_table_description(cursor, table_name)[0].name

    def get_table_description(self, cursor, table_name):
        """
        Возвращает перечень полей определенного файла
        """
        return cursor.get_file(table_name).fields


class FakeClient:
    def __init__(self, wrapper):
        pass


class DatabaseWrapper(BaseDatabaseWrapper):
    """
    Обертка на "базой данных", которая представляет из себя директорию с XSD
    """

    client_class = creation_class = features_class = ops_class = FakeClient
    introspection_class = XSDFolderIntrospection

    def cursor(self):
        return XSDFolderCursor(settings.GAR_SCHEMAS_FOLDER)



from django.contrib.postgres.indexes import (
    GinIndex,
)


class UpperGinIndex(GinIndex):
    """
    Доработка стандартного GIN индекса
    """

    def __init__(self, *, fastupdate=True, gin_pending_list_limit=None,
                 opclasses=None, **kwargs):
        kwargs['name'] = kwargs.get('name', 'base_gin_idx')
        super().__init__(
            fastupdate=fastupdate,
            gin_pending_list_limit=gin_pending_list_limit,
            opclasses=opclasses or ['gin_trgm_ops'] * len(kwargs['fields']),
            **kwargs,
        )

    def create_sql(self, model, schema_editor, using='', concurrently=False):
        """
        Позволяет оставить поддержку фильтра icontains,
        следственно - доступен поиск без учёта регистра
        Поэтому делает индекс идентичным этой операции
        """

        statement = super().create_sql(model, schema_editor, using=using)
        quote_name = statement.parts['columns'].quote_name

        statement.parts['columns'].quote_name = (
            lambda column: f'UPPER({quote_name(column)})')

        return statement



import os
import requests
from requests_toolbelt.multipart import (
    encoder,
)
from django.core.management import (
    BaseCommand,
)


class Command(BaseCommand):
    help = 'Команда публикации дампа БД в хранилище. В переменной окружение AUTH_TOKEN ожидается токен OAuth.'
    usage_str = (
        'Usage: python manage.py gar_upload_dump '
        '--path=<path_to_dump> '
        '--store_url=<url_of_store>'
        '--upload_path=<path_in_store_to_upload>'
    )

    def add_arguments(self, parser):
        super().add_arguments(parser)

        parser.add_argument(
            '--path',
            action='store',
            dest='path',
            help='Путь до файла c дампом',
            default='/home/app/web/dumps/m3_gar_dump.back',
        )
        parser.add_argument(
            '--store_url',
            action='store',
            dest='store_url',
            help='Адрес хранилища',
            default='https://cloud-api.yandex.net/v1/disk/resources',
        )
        parser.add_argument(
            '--upload_path',
            action='store',
            dest='upload_path',
            help='Путь в хранилище, кудо нужно разместить файл',
            default='gar/m3_gar_dump.back',
        )

    def handle(self, path, store_url, upload_path, **options):
        token = os.getenv('AUTH_TOKEN')
        headers = {
            'Accept': 'application/json',
            'Authorization': f'OAuth {token}',
        }
        res = requests.get(f'{store_url}/upload?path={upload_path}&overwrite=true', headers=headers).json()

        with open(path, 'rb') as f:
            data = encoder.MultipartEncoder({'file': f})
            headers['Content-Type'] = data.content_type
            try:
                requests.put(res['href'], data=data)
            except KeyError:
                print(res)



from django.conf import (
    settings,
)
from django.core.management.commands.inspectdb import (
    Command as InspectDBCommand,
)


class Command(InspectDBCommand):
    help = 'Утилита для генерации базовых моделей на основе XSD'
    usage_str = 'Usage: ./manage.py --dst <path>'

    def add_arguments(self, parser):
        super().add_arguments(parser)

        parser.add_argument(
            '--dst',
            help='Результирующий .py файл',
        )

    def handle(self, **options):
        settings.DATABASES['xsd_folder'] = {
            'ENGINE': 'm3_gar.xsd_generator',
        }
        options['database'] = 'xsd_folder'

        with open(options['dst'], 'w') as f:
            for line in self.handle_inspection(options):
                f.write(f'{line}\n')

    def get_field_type(self, connection, table_name, row):
        field_type, field_params, field_notes = super().get_field_type(
            connection, table_name, row)
        field_params['verbose_name'] = row.description

        return field_type, field_params, field_notes

    def get_meta(self, table_name, constraints, column_to_field_name, is_view, is_partition):
        result = super().get_meta(
            table_name, constraints, column_to_field_name, is_view, is_partition)

        for index, row in enumerate(result):
            if "managed = False" in row:
                result[index] = row.replace(
                    "managed = False",
                    "abstract = True",
                )
                break

        for index, row in enumerate(result):
            if "db_table" in row:
                del result[index]
                break

        return result



import datetime
import fileinput
import io
import os
import re
from io import (
    BytesIO,
)
from shutil import (
    rmtree,
)
from zipfile import (
    BadZipfile,
    ZipFile,
)

import requests
from django.conf import (
    settings,
)
from git import (
    Repo,
)
from setuptools import (
    setup,
)
from twine.commands.upload import (
    main,
)

from m3_gar.exceptions import (
    PublisherError,
)
from setup_kwargs import (
    setup_kwargs,
)


class GARSchemaManager:
    """
    Базовый класс, отвечающий за получение схем ГАР
    """

    def __init__(self, repo):
        self.repo = repo

        self.downloaded_schemas = []
        self.existing_schemas = os.listdir(settings.GAR_SCHEMAS_FOLDER)

    def process_schemas(self):
        """
        Скачивание схем с удаленного адреса
        """
        try:
            zip_archive = self._retrieve_schemas()
        except BadZipfile as e:
            raise PublisherError('Некорректный архив! {}'.format(str(e)))
        else:
            # сохраняем новые схемы с заменой
            self._save_schema(zip_archive)

    def has_schema_changes(self):
        """
        Проверяет, были ли изменения в файлах схем
        """
        # сначала добавим все файлы в репу - может, что-то пришло новое и удалилось старое
        index = self.repo.index
        index.add([
            str(settings.GAR_SCHEMAS_FOLDER / file_name)
            for file_name in os.listdir(settings.GAR_SCHEMAS_FOLDER)
        ])

        # смотрим, были ли изменения
        return bool(index.diff('HEAD'))

    def _retrieve_schemas(self):
        """
        Общий метод получения схем
        """
        raise NotImplementedError()

    def _lower_ext(self, file_name):
        """
        Отдает наименование файла с расширением в нижнем регистре
        """
        file_path, ext = os.path.splitext(file_name)
        return ''.join((file_path, ext.lower()))

    def _get_download_path(self, file_name):
        """
        Возвращает путь до файла в каталоге схем
        """
        # для консистентности всегда будем сохранять с расширением в нижнем регистре
        file_name = self._lower_ext(file_name)

        return settings.GAR_SCHEMAS_FOLDER / file_name

    def _save_schema(self, zip_archive):
        """
        Распаковывает архив и сохраняет файлы схемы
        """
        for file_name in zip_archive.namelist():
            if not self._validate_file_in_zip(file_name):
                continue

            table_name_from_xsd = self._get_table_from_xsd(file_name)
            self._delete_schema_if_found(table_name_from_xsd)

            xsd_file = zip_archive.read(file_name)
            with open(self._get_download_path(file_name), 'w+b') as write_file:
                write_file.write(xsd_file)
                self.downloaded_schemas.append(write_file.name)

    def _validate_file_in_zip(self, file_name):
        """
        Проверка, нужно ли обрабатывать файл
        """
        return True

    def _get_table_from_xsd(self, file_name):
        """
        Возвращает название таблицы из имени файла
        """
        split_file_name = file_name.split('_')
        table_name = '_'.join(split_file_name[:-6])

        return f'{table_name}_'

    def _delete_schema_if_found(self, table_name_from_xsd):
        """
        Удаление схемы из папки со схемами, если название таблицы было найдено
        в имени файла
        """
        for existing_schema in self.existing_schemas:
            if self._get_table_from_xsd(existing_schema) == table_name_from_xsd:
                file_path = settings.GAR_SCHEMAS_FOLDER / existing_schema
                os.remove(file_path)


class DownloadGARSchemaManager(GARSchemaManager):
    """
    Класс, который обеспечивает скачивание схем ГАР с указанного url
    """

    def __init__(self, url, repo, timeout=30):
        self.url = url
        self.timeout = timeout

        super().__init__(repo)

    def _retrieve_schemas(self):
        """
        Скачивание архива со схемами
        """
        response = requests.get(self.url, stream=True, timeout=self.timeout)

        return ZipFile(BytesIO(response.content))


class ArchiveGARSchemaManager(GARSchemaManager):
    """
    Класс, который обеспечивает обработку схем ГАР из уже скачанного архива
    """

    def __init__(self, path, repo):
        self.path = path

        super().__init__(repo)

    def _retrieve_schemas(self):
        """
        Отдает объект ZipFile по указанному пути
        """
        return ZipFile(self.path)

    def _get_download_path(self, file_name):
        file_name = self._lower_ext(file_name)

        return os.path.join(settings.GAR_SCHEMAS_FOLDER, file_name.split('/')[1])

    def _validate_file_in_zip(self, file_name):
        return file_name.startswith('Schemas')

    def _get_table_from_xsd(self, file_name):
        file_name = file_name.split('/')[1]

        return super()._get_table_from_xsd(file_name)


class RepoManager:
    """
    Класс для работы с репозиторием при релизе новой версии пакета
    """

    def __init__(self, test_mode=False):
        self.repo = Repo(settings.GAR_BASE_DIR.parent)
        self.test_mode = test_mode

    def index_add(self):
        """
        Добавляет новые файлы в репозиторий для последующего коммита
        """
        # файлы схем уже добавлены, добавляем миграции
        files_to_add = []
        migrations_dir = settings.GAR_BASE_DIR / 'migrations'
        for file_name in os.listdir(migrations_dir):
            if file_name.endswith('.pyc'):
                continue

            files_to_add.append(migrations_dir / file_name)

        self.repo.index.add(map(str, files_to_add))

    def save_changes(self, new_version):
        """
        Сохраняет изменения в пакете путем коммита и пуша
        """
        origin = self.repo.remote()
        push_args = [origin, self.repo.head.ref]
        if self.test_mode:
            # в этом случае создадим новую ветку по названию версии
            # и закоммитим туда, а не в мастер
            new_branch_name = f'test-patch-{new_version}'
            new_branch = self.repo.create_head(new_branch_name)
            new_branch.checkout()

            push_args = ['--set-upstream'] + push_args

        self.repo.git.commit(
            '-a',
            '-m',
            f'Автогенерация патча {new_version} по обновлению формата ГАР',
        )
        self.repo.git.push(*push_args)


class ReleasePreparer:
    """
    Класс, ответственный за подготовку к выпуску.
    Повышает версию пакета и дополняет чейнжлог автосгенерированной фразой
    """

    def __init__(self):
        self.version_file = str(settings.GAR_BASE_DIR / 'version.py')
        self.changelog_file = str(settings.GAR_BASE_DIR.parent / 'CHANGELOG.rst')
        self.new_version_str = ''

    def prepare(self):
        """
        Основной метод класса подготовки
        """
        self._update_version()
        self._update_changelog()

    def _update_version(self):
        """
        Увеличивает последнюю цифру в файле версии
        """
        with fileinput.input(self.version_file, inplace=True) as version_file:
            for line in version_file:
                if line.startswith('VERSION = '):
                    version_str = line.split('=')[1].strip()[1:-1]
                    major, minor, patch = map(int, version_str.split(','))
                    next_patch = patch + 1

                    self.new_version_str = f'{major}.{minor}.{next_patch}'

                    print(f'VERSION = ({major}, {minor}, {next_patch})')

                else:
                    print(line.strip('\n'))

    def _update_changelog(self):
        """
        Создает новую запись в чейнжлоге о текущей версии
        """
        if not self.new_version_str:
            return

        today = datetime.date.today()
        today_str = today.strftime('%Y.%m.%d')

        with io.open(self.changelog_file, 'r+', encoding='utf-8') as changelog_file:
            content = changelog_file.read()
            changelog_file.seek(0, 0)

            changelog_file.write(f'{today_str} v.{self.new_version_str}\n')
            changelog_file.write('\n')
            changelog_file.write('* Автогенерируемый патч по обновлению формата ГАР\n')
            changelog_file.write('\n')

            changelog_file.write(content)


class Releaser:
    """
    Класс, отвечающий за создание и выпуск пакета
    """

    def __init__(self, new_version, test_mode=False):
        self.new_version = new_version
        self.test_mode = test_mode

    def release(self):
        """
        Основной метод запуска сборки
        """
        self._build_dist()

        try:
            self._publish()
        finally:
            self._clear_build_folder()

    def _build_dist(self):
        """
        Создание пакета
        """
        new_setup_kwargs = setup_kwargs.copy()
        new_setup_kwargs.update({
            'version': self.new_version,
            'script_name': 'setup.py',
            'script_args': ['sdist', 'bdist_wheel'],
        })

        cwd = os.getcwd()
        os.chdir(settings.BASE_DIR.parent)

        try:
            setup(**new_setup_kwargs)
        finally:
            os.chdir(cwd)

    def _clear_build_folder(self):
        """
        Очистка после билда пакета
        """
        rmtree(settings.BASE_DIR / 'build')
        rmtree(settings.BASE_DIR / 'dist')

    def _publish(self):
        """
        Загрузка пакета на pypi
        Для получения логина-пароля используется файл ~/.pypirc, там должен
        быть указан адрес до pypi под категорией pypi,
        и адрес до тестового pypi под категорией testpypi, если нужно
        использовать тестовый режим скрипта
        """
        pypirc_config = ['-r', 'pypi']
        if self.test_mode:
            pypirc_config = ['-r', 'testpypi']

        args = pypirc_config + ['dist/*']

        cwd = os.getcwd()
        os.chdir(settings.BASE_DIR.parent)

        try:
            main(args)
        finally:
            os.chdir(cwd)



from m3_gar.base_models import (
    Apartments as BaseApartments,
    ApartmentTypes as BaseApartmentTypes,
    OperationTypes,
)
from m3_gar.models.reestr import (
    ReestrObjects,
)
from m3_gar.models.util import (
    RegionCodeModelMixin,
    add_params,
    make_fk,
)


__all__ = ['Apartments', 'ApartmentTypes']


class ApartmentTypes(BaseApartmentTypes):
    """
    Сведения по типам помещений
    """
    class Meta:
        verbose_name = 'Тип помещения'
        verbose_name_plural = 'Типы помещений'


class Apartments(BaseApartments, RegionCodeModelMixin):
    """
    Сведения по помещениям
    """

    level = 11

    class Meta:
        verbose_name = 'Помещение'
        verbose_name_plural = 'Помещения'


make_fk(Apartments, 'aparttype', to=ApartmentTypes, null=True, blank=True)
make_fk(Apartments, 'opertypeid', to=OperationTypes)
make_fk(Apartments, 'objectid', to=ReestrObjects)

add_params(Apartments, 'm3_gar.ApartmentsParams')



# Generated by Django 3.2.4 on 2021-07-07 13:52

import django.db.models.deletion
from django.db import (
    migrations,
    models,
)


class Migration(migrations.Migration):

    dependencies = [
        ('m3_gar', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='admhierarchy',
            name='objectid',
            field=models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.reestrobjects', verbose_name='Глобальный уникальный идентификатор объекта'),
        ),
        migrations.AlterField(
            model_name='munhierarchy',
            name='objectid',
            field=models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.reestrobjects', verbose_name='Глобальный уникальный идентификатор адресного объекта '),
        ),
    ]



import inspect
import re
from importlib import (
    import_module,
)

from django.conf import (
    settings,
)
from django.core.exceptions import (
    ImproperlyConfigured,
)
from django.db import (
    models,
)

import m3_gar.base_models
import m3_gar.models


def get_models():
    """
    Возвращает список моделей, относящихся к ГАР
    """
    base_models = tuple(
        obj for name, obj
        in inspect.getmembers(m3_gar.base_models)
        if inspect.isclass(obj) and issubclass(obj, models.Model)
    )

    concrete_models = [
        obj for name, obj
        in inspect.getmembers(m3_gar.models)
        if inspect.isclass(obj) and issubclass(obj, base_models)
    ]

    return concrete_models


def get_table_names_from_models():
    """
    Возвращает названия таблиц
    """

    table_names = [
        re.sub('(?!^)([A-Z]+)', r'_\1', model.__name__).lower()
        for model in get_models()
    ]

    return tuple(table_names)


def get_model_from_table_name(table_name):
    model_name = ''.join(x.title() for x in table_name.split('_'))
    model = getattr(m3_gar.models, model_name, None)
    return model


def get_table_row_filters():
    """
    Перенос функционала из config в функцию
    Достает фильтры для таблиц

    Оригинальное описание:
    см. m3_gar.importer.filters
    указывается список путей к функциям-фильтрам
    фильтры применяются к *каждому* объекту
    один за другим, пока не закончатся,
    либо пока какой-нибудь из них не вернёт None
    если фильтр вернул None, объект не импортируется в БД

    пример:

    GAR_TABLE_ROW_FILTERS = {
        'addrobj': (
            'm3_gar.importer.filters.example_filter_yaroslavl_region',
        ),
        'house': (
            'm3_gar.importer.filters.example_filter_yaroslavl_region',
        ),
    }
    """
    row_filters = getattr(settings, 'GAR_TABLE_ROW_FILTERS', {})
    table_row_filters = {}

    for flt_table, flt_list in row_filters.items():
        if flt_table in get_table_names_from_models():
            for flt_path in flt_list:
                try:
                    module_name, _, func_name = flt_path.rpartition('.')
                    flt_module = import_module(module_name)
                    flt_func = getattr(flt_module, func_name)
                except (ImportError, AttributeError):
                    raise ImproperlyConfigured(
                        'Table row filter module `{0}` does not exists'.format(
                            flt_path))
                else:
                    table_row_filters.setdefault(flt_table, []).append(flt_func)

    return table_row_filters



from django.contrib.postgres.indexes import (
    HashIndex,
)

from m3_gar.base_models import (
    ObjectLevels as BaseObjectLevels,
    ReestrObjects as BaseReestrObjects,
)
from m3_gar.models.util import (
    RegionCodeModelMixin,
    make_fk,
)


__all__ = ['ObjectLevels', 'ReestrObjects']


class ObjectLevels(BaseObjectLevels):
    """
    Сведения по уровням адресных объектов
    """
    class Meta:
        verbose_name = 'Уровень адресного объекта'
        verbose_name_plural = 'Уровни адресных объектов'

    @staticmethod
    def get_model_by_level(level):
        from m3_gar.models import (
            AddrObj,
            Apartments,
            Carplaces,
            Houses,
            Rooms,
            Steads,
        )

        known_models = [
            Apartments,
            Carplaces,
            Houses,
            Rooms,
            Steads,
        ]
        default_model = AddrObj

        level_map = {
            model.level: model
            for model in known_models
        }

        model = level_map.get(level, default_model)

        return model

    @property
    def addr_obj_model(self):
        return self.get_model_by_level(self.level)


class ReestrObjects(BaseReestrObjects, RegionCodeModelMixin):
    """
    Сведения об адресном элементе в части его идентификаторов
    """
    class Meta:
        verbose_name = 'Идентификатор адресного элемента'
        verbose_name_plural = 'Идентификаторы адресных элементов'

        indexes = (
            HashIndex(
                fields=('objectguid',),
            ),
        )

    @property
    def addr_obj_model(self):
        levels_model = self._meta.get_field('levelid').target_field.model
        return levels_model.get_model_by_level(self.levelid_id)

    @property
    def addr_obj_set(self):
        addr_obj_model = self.addr_obj_model

        class AddrObjManager(addr_obj_model._default_manager.__class__):
            def get_queryset(manager):
                return super().get_queryset().filter(objectid=self.objectid)

        manager = AddrObjManager()
        manager.model = addr_obj_model

        return manager


make_fk(ReestrObjects, 'levelid', to=ObjectLevels)



# Generated by Django 3.2.4 on 2021-10-13 11:06

import django.db.models.deletion
from django.db import (
    migrations,
    models,
)


class Migration(migrations.Migration):

    dependencies = [
        ('m3_gar', '0003_auto_20210906_0744'),
    ]

    operations = [
        migrations.CreateModel(
            name='AddrObjParams',
            fields=[
                ('id', models.BigIntegerField(primary_key=True, serialize=False, verbose_name='Идентификатор записи')),
                ('objectid', models.BigIntegerField(verbose_name='Глобальный уникальный идентификатор адресного объекта ')),
                ('changeid', models.BigIntegerField(blank=True, null=True, verbose_name='ID изменившей транзакции')),
                ('changeidend', models.BigIntegerField(verbose_name='ID завершившей транзакции')),
                ('value', models.CharField(max_length=8000, verbose_name='Значение параметра')),
                ('updatedate', models.DateField(verbose_name='Дата внесения (обновления) записи')),
                ('startdate', models.DateField(verbose_name='Дата начала действия записи')),
                ('enddate', models.DateField(verbose_name='Дата окончания действия записи')),
                ('region_code', models.SmallIntegerField(verbose_name='Код региона')),
                ('typeid', models.ForeignKey(db_column='typeid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.paramtypes', verbose_name='Тип параметра')),
            ],
            options={
                'verbose_name': 'Параметр',
                'verbose_name_plural': 'Параметры',
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ApartmentsParams',
            fields=[
                ('id', models.BigIntegerField(primary_key=True, serialize=False, verbose_name='Идентификатор записи')),
                ('objectid', models.BigIntegerField(verbose_name='Глобальный уникальный идентификатор адресного объекта ')),
                ('changeid', models.BigIntegerField(blank=True, null=True, verbose_name='ID изменившей транзакции')),
                ('changeidend', models.BigIntegerField(verbose_name='ID завершившей транзакции')),
                ('value', models.CharField(max_length=8000, verbose_name='Значение параметра')),
                ('updatedate', models.DateField(verbose_name='Дата внесения (обновления) записи')),
                ('startdate', models.DateField(verbose_name='Дата начала действия записи')),
                ('enddate', models.DateField(verbose_name='Дата окончания действия записи')),
                ('region_code', models.SmallIntegerField(verbose_name='Код региона')),
                ('typeid', models.ForeignKey(db_column='typeid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.paramtypes', verbose_name='Тип параметра')),
            ],
            options={
                'verbose_name': 'Параметр',
                'verbose_name_plural': 'Параметры',
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='CarplacesParams',
            fields=[
                ('id', models.BigIntegerField(primary_key=True, serialize=False, verbose_name='Идентификатор записи')),
                ('objectid', models.BigIntegerField(verbose_name='Глобальный уникальный идентификатор адресного объекта ')),
                ('changeid', models.BigIntegerField(blank=True, null=True, verbose_name='ID изменившей транзакции')),
                ('changeidend', models.BigIntegerField(verbose_name='ID завершившей транзакции')),
                ('value', models.CharField(max_length=8000, verbose_name='Значение параметра')),
                ('updatedate', models.DateField(verbose_name='Дата внесения (обновления) записи')),
                ('startdate', models.DateField(verbose_name='Дата начала действия записи')),
                ('enddate', models.DateField(verbose_name='Дата окончания действия записи')),
                ('region_code', models.SmallIntegerField(verbose_name='Код региона')),
                ('typeid', models.ForeignKey(db_column='typeid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.paramtypes', verbose_name='Тип параметра')),
            ],
            options={
                'verbose_name': 'Параметр',
                'verbose_name_plural': 'Параметры',
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='HousesParams',
            fields=[
                ('id', models.BigIntegerField(primary_key=True, serialize=False, verbose_name='Идентификатор записи')),
                ('objectid', models.BigIntegerField(verbose_name='Глобальный уникальный идентификатор адресного объекта ')),
                ('changeid', models.BigIntegerField(blank=True, null=True, verbose_name='ID изменившей транзакции')),
                ('changeidend', models.BigIntegerField(verbose_name='ID завершившей транзакции')),
                ('value', models.CharField(max_length=8000, verbose_name='Значение параметра')),
                ('updatedate', models.DateField(verbose_name='Дата внесения (обновления) записи')),
                ('startdate', models.DateField(verbose_name='Дата начала действия записи')),
                ('enddate', models.DateField(verbose_name='Дата окончания действия записи')),
                ('region_code', models.SmallIntegerField(verbose_name='Код региона')),
                ('typeid', models.ForeignKey(db_column='typeid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.paramtypes', verbose_name='Тип параметра')),
            ],
            options={
                'verbose_name': 'Параметр',
                'verbose_name_plural': 'Параметры',
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='RoomsParams',
            fields=[
                ('id', models.BigIntegerField(primary_key=True, serialize=False, verbose_name='Идентификатор записи')),
                ('objectid', models.BigIntegerField(verbose_name='Глобальный уникальный идентификатор адресного объекта ')),
                ('changeid', models.BigIntegerField(blank=True, null=True, verbose_name='ID изменившей транзакции')),
                ('changeidend', models.BigIntegerField(verbose_name='ID завершившей транзакции')),
                ('value', models.CharField(max_length=8000, verbose_name='Значение параметра')),
                ('updatedate', models.DateField(verbose_name='Дата внесения (обновления) записи')),
                ('startdate', models.DateField(verbose_name='Дата начала действия записи')),
                ('enddate', models.DateField(verbose_name='Дата окончания действия записи')),
                ('region_code', models.SmallIntegerField(verbose_name='Код региона')),
                ('typeid', models.ForeignKey(db_column='typeid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.paramtypes', verbose_name='Тип параметра')),
            ],
            options={
                'verbose_name': 'Параметр',
                'verbose_name_plural': 'Параметры',
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SteadsParams',
            fields=[
                ('id', models.BigIntegerField(primary_key=True, serialize=False, verbose_name='Идентификатор записи')),
                ('objectid', models.BigIntegerField(verbose_name='Глобальный уникальный идентификатор адресного объекта ')),
                ('changeid', models.BigIntegerField(blank=True, null=True, verbose_name='ID изменившей транзакции')),
                ('changeidend', models.BigIntegerField(verbose_name='ID завершившей транзакции')),
                ('value', models.CharField(max_length=8000, verbose_name='Значение параметра')),
                ('updatedate', models.DateField(verbose_name='Дата внесения (обновления) записи')),
                ('startdate', models.DateField(verbose_name='Дата начала действия записи')),
                ('enddate', models.DateField(verbose_name='Дата окончания действия записи')),
                ('region_code', models.SmallIntegerField(verbose_name='Код региона')),
                ('typeid', models.ForeignKey(db_column='typeid', on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='m3_gar.paramtypes', verbose_name='Тип параметра')),
            ],
            options={
                'verbose_name': 'Параметр',
                'verbose_name_plural': 'Параметры',
                'abstract': False,
            },
        ),
        migrations.DeleteModel(
            name='Param',
        ),
    ]
